/* ---------------------------------------------------------------------------
   eibdrv.c
   ---------------------------------------------------------------------------
   eibdrv Version 0.2.1
   Copyright (C) 2002, Wolfgang Tumfart
                       Donaustrasse 104/9
                       A-2344 Maria Enzersdorf
                       Austria (Europe)
                       tumfart@auto.tuwien.ac.at

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   675 Mass Ave, Cambridge, MA 02139, USA.
   --------------------------------------------------------------------------- */


#ifndef __KERNEL__
#  define __KERNEL__
#endif

#ifndef MODULE
#  define MODULE
#endif

#define VERSION_CODE(vers,rel,seq) (((vers)<<16) | ((rel)<<8) | (seq))

// define FTSTATION_BCU2 if module is to be used
// as BCU2 for test purposes

#ifdef FTSTATION_BCU2
#   define FT_DIR 1
#else
#   define FT_DIR 0
#endif


#include <linux/module.h>
#include <linux/malloc.h>
#include <linux/serial.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/tqueue.h>
#include <linux/ioctl.h>
#include <linux/delay.h>

#include <asm/system.h>
#include <asm/fcntl.h>

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
#  include <linux/poll.h>
#  include <asm/uaccess.h>
#else 
#  include <asm/segment.h>
#endif

#include "eibdrv.h"

/*  --------------------------------------------------------------------------- */

#define EIBDRV_VERSION "0.2.1"
#define EIBDRV_VERSION_CODE 513

// Internal buffers/lists

#define FT_PROC_BLOCKED    0x01        // write request of type acknowledged
#define FT_PROC_NONBLOCKED 0x02        // write request of type non acknowledged

#define FT_RMPROC_AUTO     0x01        // remove by server process
#define FT_RMPROC_MAN      0x02        // remove by application process
#define FT_PROC_FREE       0x04        // write request has been processed and
                                       // can be removed from the list

#define FT_PROC_OPEN 0x08              // ==> rproc->flags
                                       // indicates that a slot of the read
                                       // process list is in use

// FT1.2-protocol parameter

#define FT_EXCHANGE_TIMEOUT 510        // time between request and ack (in bit)
#define FT_LINE_IDLE_TIMEOUT 33        // time between two frames (in bit)
#define FT_REPEAT_LIMIT 3              // number of transimissions of an 
                                       // unacknowledged frame

#define FT_FIXED_LENGTH_MSG 0x10       // start character of frame w fixed length
#define FT_VAR_LENGTH_MSG   0x68       // start character of frame w variable length
#define FT_ACK              0xE5       // acknowledge frame

#define FT_FCB_VALID 0x01              // fcb valid/void depending on 
#define FT_FCB_VOID  0x00              // frame type

#define FT_STATE_INIT                  0   
#define FT_STATE_READY                 2    
#define FT_STATE_WAIT_FOR_UDATA_ACK    3
#define FT_STATE_WAIT_FOR_RESP_STATUS  4
#define FT_STATE_WAIT_FOR_RESET_ACK    5
#define FT_STATE_SHUTDOWN              9

// Module method: Open, Release

#define FT_SHARED_RW    0x00           // ==> ftstation.f_exclusive
#define FT_EXCLUSIVE_RW 0x01           // indicates whether the FT1.2-
                                       // protocoll station is opened
                                       // for exclusive or shared
                                       // Read/Write acess.


/*  --------------------------------------------------------------------------- */

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
static mm_segment_t fs;
#else
unsigned long fs;
#  ifdef __ARCH_I386_POSIX_TYPES_H
typedef unsigned long suseconds_t;
#  endif
#endif

#if defined(__alpha__) 
#  define count_t unsigned long
#  define read_write_t long
#else
#  define count_t int
#  define read_write_t int
#endif

int c=0;                               // Numbering of messages ensures, that
                                       // all messages are display on the console

/*  =========================================================================== */
/*  FT-protocol station                                                         */
/*  =========================================================================== */

char *port="/dev/ttyS1";  
unsigned int major=100;   
int mode=FT_SERVERMODE; 

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
MODULE_PARM(port,"s");
MODULE_PARM(major,"i");
MODULE_PARM(mode,"i");
#endif

int baudrate=19200;        
int symb_baudrate=B19200;  


// lock variable for common ressources
typedef struct sv { int               val;             
                    struct wait_queue *wq;
                  } sv;
                  
sv sv_ser_write={1,NULL};         // writing lock of serial interface
sv sv_ser_read ={1,NULL};         // reading lock of serial interface
sv sv_ready    ={1,NULL};         // Initiating a new transmission cycle

// Serial port used by the eibdrv driver to communicate via the
// BCU2 with the EIB bus.

typedef struct SerPort
   { struct file       *filp;
     struct termios    old_termios;
     struct tty_struct *tty;
     struct inode      *inodp;
   } SerPort;


// Structure containing all internal state variables and buffers of the 
// protocol station, that handles the communication with the BCU2 via 
// the FT1.2 protocol.

typedef struct Eib_wpbuf Eib_wpbuf;
typedef struct Eib_rpbuf Eib_rpbuf;
typedef struct Eib_buf   Eib_buf;

typedef struct Eib_FTStation
   { int               state;               // state of the protocol automata
     int 	       servproc;            // server processes opened the device
     int               servfunc;            // server processes working 
     int               f_exclusive;         // device opened exclusively
     int               exchange_timeout;    // max delay between request and answer (in jiffies)
     struct timeval    exchange_time;       // max delay between request and answer
     time_t            line_idle_timeout_sec;    // min delay betw to msg (in sec)
     suseconds_t       line_idle_timeout_usec;   // min delay betw two msg (in usec)
     struct timeval    last_seract;         // timestamp of last write/read access to tty
     struct timeval    start_write;         // timestamp of beginn of write
     struct timeval    end_write;           // timestamp of end of write w/o retransmissions
     int               fcb;                 // fcb of next write message
     int               rec_fcb;             // fcb of last received message
     int               wrepeat;             // number of repetitions of write message
     int               rrepeat;             // number of retransmissions of read message
     SerPort           port;                // serial port
     Eib_statistics    stat;	            // statistic information
     Eib_wpbuf         *wpbuf;              // buffer for write accesses
     Eib_rpbuf         *rpbuf;              // buffer for read accesses & received messges
     Eib_buf           *outbuf;             // buffer for write messages
     struct fasync_struct  *pasync;         // processes to be notified with SIGIO
   } Eib_FTStation;



static Eib_FTStation ftstation;             // protocol station


struct wait_queue *wq_wproc=NULL;           // blocking write processes in acknowledge                                             // mode waiting for the result of 
                                            // write access
struct wait_queue *wq_rproc=NULL;           // blocking read processes waiting for a
                                            // message from BCU2 in their buffer
struct wait_queue *wq_fullwproc=NULL;       // blocking write processes in standard
                                            // mode waiting for an empty slot in
                                            // outbuf and wpbuf

struct wait_queue *wq_ser_write=NULL;       // serial interface write process
                                            // waits for new data in outbuf
struct wait_queue *wq_wait_for_ack=NULL;    // ser_write waits for ack or
                                            // exchange timeout                                       

struct wait_queue *wq_shutdown=NULL;        // cleanup_module waits for idle_read
                                            // and idle_write to notify their closure
                                            // by decrementing ftstation.servproc 

struct wait_queue *sleep_queue=NULL;
/*  =========================================================================== */
/*  Message buffer                                                              */
/*  =========================================================================== */

    /*  If the module is installed in server mode, buffers are used for both 
        outgoing and incoming messages. They are implemented as ring buffer 
        with a maximum message length of EIB_BUF_ELEMENT_SIZE unsigned 
        characters.
     */
    
struct Eib_buf 
   { unsigned char *mem;               // start of memory region for slots
     unsigned char *next_read;         // slot containing next message to be read
     unsigned char *next_write;        // slot for next message to be written
     unsigned long count;              // number of messages in buffer
     unsigned long size;               // number of available slots in buffer
   };
   

    /*  This function allocates the memory for a message buffer of size elements, 
        each with a maximum length of EIB_BUF_ELEMENT_SIZE characters, according
        to the maximum message length specified for the FT1.2-protocol in the 
        EIB system.
    
        The function returns a pointer to the buffer, which is used as parameter 
        for all other function calls regarding the message buffer, or NULL, if 
        the necessary memory could not be allocated.
     */

Eib_buf *eib_get_buf(unsigned long size);


    /*  This function deallocates the memory used for a message buffer and has to 
        be called if the buffer is not longer to be used.
     */

void eib_return_buf(Eib_buf *buf);


    /*  This function returns the oldest unread message in the buffer; the read 
        pointer is set to the next unread message. The argument element points 
        to a preallocated array of type unsigned character, into which the message 
        is copied. It is safe to choose the size of the array not shorter than 
        EIB_BUF_ELEMENT_SIZE. The length of the message is returned in the 
        function value.
     */

int eib_read_buf(Eib_buf *buf,unsigned char *element);


    /*  This function copies len characters (maximum is EIB_BUF_ELEMENT_SIZE) from 
        element into a slot of the message buffer. The function returns 
        -ENOBUFS, if all slots of the message buffer are filled, or 0 otherwise.
     */

int eib_write_buf(Eib_buf *buf,unsigned char *element,int len);


    /*  This function resizes the buffer to new_size. It returns -ENOMEM, if 
        memory cannot be allocated for the new buffer; -ESIZE, if the new list 
        is too small to hold all entries of the old list and resizing is 
        therefore rejected; or 0 otherwise.
     */
     
int eib_resize_buf(Eib_buf **buf,unsigned long new_size);


unsigned long eib_get_buf_count(Eib_buf *buf);
unsigned long eib_get_buf_size(Eib_buf *buf);
void eib_display_buf(Eib_buf *buf);

// for internal use only
void eib_init_buf(Eib_buf *buf);
void eib_calc_nextpos_buf(Eib_buf *buf,unsigned char **nextpos);
void eib_copymem_buf(Eib_buf *old_buf,Eib_buf *new_buf);


/*  =========================================================================== */
/*  List of write processes                                                     */
/*  =========================================================================== */

/*  References to all write processes (blocking and nonblocking) are kept in 
    an ordered list until the write access has been achknowledged by the BCU2 
    or the write access has been unsuccessfull for a certain number of 
    repetitions.
 */

unsigned long wpbuf_size=20;          
unsigned long outbuf_size=10;     

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
MODULE_PARM(wpbuf_size,"l");
MODULE_PARM(outbuf_size,"l");
#endif

typedef struct Eib_wproc
   { pid_t pid;                        // id of write process 
     int flags;                        // remove and processed-flag, 0 if empty slot
     int result;                       // -EAGAIN if write unsuccessful, 0..23 else
     unsigned long next;               // index 1..wpbuf_size of next process
     unsigned long prev;               // index 1.. wpbuf_size of previous process
   } Eib_wproc;
   
struct Eib_wpbuf
   { struct Eib_wproc *mem;            // start of memory region for slots
     unsigned long first;              // index 1..wpbuf_size of first slot
     unsigned long last;               // index 1..wpbuf_size of last slot
     unsigned long next_result;        // index of slot, where next result to write
     unsigned long count;              // number of processes in list
     unsigned long size;               // number of available slots in buffer
   };
   

    /*  This function allocates memory for a process list of size elements (slots).
        The function returns a pointer to the buffer, which is used as parameter 
        for all other function calls regarding the process list, or NULL, if the 
        necessary memory could not be allocated.
     */

Eib_wpbuf *eib_get_wpbuf(unsigned long size);


    /*  This function deallocates the memory used for a process list and has to be 
        called if the process list is not longer to be used.
     */

void eib_return_wpbuf(Eib_wpbuf *buf);


    /*  This function sets a process with the process-id pid into the process 
        list. flags indicates whether the process is in blocking (FT_PROC_BLOCKED) 
        or nonblocking (FT_PROC_NONBLOCKED) write mode.
    
        The function returns -ENOBUFS, if no empty slot in the process list is
        available, or 0 otherwise.
     */
    
int eib_write_wpbuf(Eib_wpbuf *buf,pid_t pid,int flags);


    /*  This function returns the result of a write process and removes it from 
        the list. The function returns -ENOPROC, if the process with the given 
        id is not in the process list; -ENORESULT, if the write access has not 
        been processed yet and can therefore not be removed from the list; or
        the number of written and acknowledged characters by the BCU2.
     */

int eib_condrm_wpbuf(Eib_wpbuf *buf,pid_t pid);


    /*  This function stores the result of a write access in the appropriate slot 
        in the process list.
     */
  
void eib_result_wpbuf(Eib_wpbuf *buf,int result);


    /*  This function resizes the process list to new_size. It returns -ENOMEM, 
        if memory cannot be allocated for the new list; -ESIZE, if the new list 
        is too small to hold all entries of the old list, and resizing is 
        therefore rejected; or 0 otherwise.
     */
 
int eib_resize_wpbuf(Eib_wpbuf **buf,unsigned long new_size);

 
    /*  This function counts the occurences of processes with process-ID pid
        in the process list.
     */
     
unsigned long eib_countinst_wpbuf(Eib_wpbuf *buf,pid_t pid);


unsigned long eib_get_wpbuf_count(Eib_wpbuf *buf);
unsigned long eib_get_wpbuf_size (Eib_wpbuf *buf);
void eib_display_wpbuf(Eib_wpbuf *buf);

// for internal use only
void eib_init_wpbuf(Eib_wpbuf *buf);


/*  =========================================================================== */
/*  List of read processes                                                      */
/*  =========================================================================== */

/*  A reference to all processes opening the device is kept in an unordered 
    list, until they close the device. Each process has an individual message 
    buffer, where all unread messages received from the BCU2 are stored.
 */

unsigned long rpbuf_size=30;            
unsigned long rpbuf_msgbuf_size=10;    

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
MODULE_PARM(rpbuf_size,"l");
MODULE_PARM(rpbuf_msgbuf_size,"l");
#endif

typedef struct Eib_rproc
   { struct file *filp;                // open device
     int flags;                        // type, in use
     int wmode;                        // Standard/Acknowledged Write Mode
     Eib_buf *msgbuf;                  // message buffer
   } Eib_rproc;

struct Eib_rpbuf
   { struct Eib_rproc *mem;            // start of memory region for slots
     unsigned long count;              // number of open devices
     unsigned long size;               // number of available slots
     unsigned long msgbuf_size;        // number of messages per open device
   };
   

    /*  This function allocates memory for a process list of size elements 
        (slots), which can store a maximum of msgbuf_size messages for each
        process.
    
        The function returns a pointer to the buffer, which is used as parameter 
        for all other function calls regarding the process list, or NULL, if the 
        necessary memory could not be allocated.
     */

Eib_rpbuf *eib_get_rpbuf(unsigned long size,unsigned long msgbuf_size);


    /*  This function deallocates the memory used for a process list and has to be 
        called if the process list is not longer to be used.
     */

void eib_return_buf(Eib_buf *buf);


    /*  This function inserts a process into the first free slot of the process 
        list. flags indicates whether the device is in blocking (FT_PROC_BLOCKED) 
        or nonblocking (FT_PROC_NONBLOCKED) read mode.
        
        The function returns -ENOBUFS, if no empty slot in the process list is
        available, -EDEVISOPEN, if the device has already been opened by this 
        process or 0 otherwise.
     */

int eib_ins_rpbuf(Eib_rpbuf *buf,struct file *filp,int flags);


     /*  This function removes an open process from the process list. 
        
         The function return -ENOPROC, if the process is not kept in the process
         list or 0 otherwise.
      */
      
int eib_rm_rpbuf(Eib_rpbuf *buf,struct file *filp);


    /*  This function copies len characters (maximum is EIB_BUF_ELEMENT_SIZE)
        from element into the message buffers of all processes in the process
        list, if these buffers have empty message slots. 
         
        The function returns -ENOBUFS, if at least one message buffer is full
        - although the message is copied into all other message buffers - or
        0 otherwise.
     */ 

int eib_wmsg_rpbuf(Eib_rpbuf *buf,unsigned char *element,int len);


    /*  This function returns the oldest unread message from the message buffer
        of a process. The argument element points to a preallocated array of 
        type unsigned character, into which the message is copied. It is safe 
        to choose the size of the array not shorter than EIB_BUF_ELEMENT_SIZE. 
        
        The functions returns the length of the message or 0, if there is no
        message available in the buffer or the process is not kept in the 
        process list.
     */

int eib_rmsg_rpbuf(Eib_rpbuf *buf,struct file *filp,unsigned char *element);


    /*  This function resizes the proces list to new_size and/or the message
        buffers of the processes to new_msgbuf_size. It returns -ENOMEM, if 
        memory cannot be allocated for the new list; -ESIZE, if the new list 
        is too small to hold all entries of the old lists and the resizing is 
        therefore rejected; or 0 otherwise.
     */

int eib_resize_rpbuf(Eib_rpbuf **buf,unsigned long new_size,unsigned long new_msgbuf_size);


    /*  This function returns the number of messages in the buffer for a
        process or -ENOPROC, if the process is not kept in the process list.
     */
     
unsigned long eib_get_rpbuf_msgcount(Eib_rpbuf *buf,struct file *filp);
int eib_set_rpbuf_wmode(Eib_rpbuf *buf,struct file *filp,int wmode);
int eib_get_rpbuf_wmode(Eib_rpbuf *buf,struct file *filp);
void eib_display_rpbuf(Eib_rpbuf *buf);

unsigned long eib_get_rpbuf_count(Eib_rpbuf *buf);
unsigned long eib_get_rpbuf_size(Eib_rpbuf *buf);
unsigned long eib_get_rpbuf_msgbuf_size(Eib_rpbuf *buf);

// for internal use only
void eib_init_rpbuf(Eib_rpbuf *pbuf);


/*  =========================================================================== */

Eib_buf *eib_get_buf(unsigned long size)
{  Eib_buf *buf;

   buf=(Eib_buf *)kmalloc(sizeof(Eib_buf)+size*(EIB_BUF_ELEMENT_SIZE+1)*
                    sizeof(unsigned char),GFP_KERNEL);
   if (buf!=NULL)
   {  buf->mem=(unsigned char *)(buf+1);
      buf->size=size;
      eib_init_buf(buf);
   }
   return buf;
}

void eib_return_buf(Eib_buf *buf) 
{  kfree(buf); 
}

int eib_read_buf(Eib_buf *buf,unsigned char *element)
{  int i,len;

   if (buf->next_read==NULL) return 0;

   len=(int)*buf->next_read;
   for (i=0;i<len;i++)
      *(element+i)=*(buf->next_read+1+i);
   buf->count--;
   if (buf->count>0)
      eib_calc_nextpos_buf(buf,&buf->next_read);
   else 
      buf->next_read=NULL;
   
   return len;
}

int eib_write_buf(Eib_buf *buf,unsigned char *element,int len)
{  int i;

   if (buf->count==buf->size) return -ENOBUFS;
   
   if (len>EIB_BUF_ELEMENT_SIZE)   
      len=EIB_BUF_ELEMENT_SIZE;
      
   *buf->next_write=(unsigned char)len;
   for (i=0;i<len;i++)
      *(buf->next_write+1+i)=*(element+i);
   buf->count++;
   if (buf->next_read==NULL)
      buf->next_read=buf->next_write;
   eib_calc_nextpos_buf(buf,&buf->next_write);
   
   return 0;
}

void eib_copymem_buf(Eib_buf *old_buf,Eib_buf *new_buf)
{  unsigned long diff_size;
   
   diff_size=new_buf->size-old_buf->size;
   if (old_buf->next_write>old_buf->next_read)
   {  memcpy(new_buf->mem,old_buf->next_read,
         (old_buf->next_write-old_buf->next_read)*sizeof(unsigned char));
      new_buf->next_read=new_buf->mem;
      new_buf->next_write=new_buf->mem+(old_buf->next_write-old_buf->next_read);
   }
   else  // old_buf->next_write <= old_buf->next_read
   {  memcpy(new_buf->mem,old_buf->mem,
         (old_buf->next_write-old_buf->mem)*sizeof(unsigned char));
      memcpy(new_buf->mem+(old_buf->next_read-old_buf->mem+
         diff_size*(EIB_BUF_ELEMENT_SIZE+1)*sizeof(unsigned char)),
         old_buf->next_read,
         (old_buf->mem+old_buf->size*(EIB_BUF_ELEMENT_SIZE+1)*sizeof(unsigned char))-(old_buf->next_read));
      new_buf->next_read=
         new_buf->mem+(old_buf->next_read-old_buf->mem+(diff_size*(EIB_BUF_ELEMENT_SIZE+1)*sizeof(unsigned char)));
      new_buf->next_write=new_buf->mem+(old_buf->next_write-old_buf->mem);
   }
   new_buf->count=old_buf->count;
}

int eib_resize_buf(Eib_buf **buf,unsigned long new_size)
{  Eib_buf *old_buf,*new_buf;

   if ((new_buf=eib_get_buf(new_size))==NULL) return -ENOMEM;
   old_buf=(*buf);
   if (new_size==old_buf->size) 
   {  eib_return_buf(new_buf);
      return 0;
   }
   if ((new_size<old_buf->count) || (new_size==0)) 
   {  eib_return_buf(new_buf);
      return -ESIZE;
   }
   
   if (old_buf->count>0)
      eib_copymem_buf(old_buf,new_buf);
   eib_return_buf(*buf);
   *buf=new_buf;
   return 0;
}

void eib_calc_nextpos_buf(Eib_buf *buf,unsigned char **nextpos)
{  
   if (*nextpos == buf->mem+(buf->size-1)*(EIB_BUF_ELEMENT_SIZE+1)*sizeof(unsigned char))
      *nextpos=buf->mem;
   else
      *nextpos+=(EIB_BUF_ELEMENT_SIZE+1);
}

unsigned long eib_get_buf_count(Eib_buf *buf) 
{  return buf->count;  
}

unsigned long eib_get_buf_size(Eib_buf *buf) 
{  return buf->size;  
}

void eib_init_buf(Eib_buf *buf)
{  buf->next_read=NULL;
   buf->next_write=buf->mem;
   buf->count=0;
}

void eib_display_buf(Eib_buf *buf)
{  int i,k,len;
   unsigned long old_count;
   unsigned char *old_next_read;
   unsigned char text[EIB_BUF_ELEMENT_SIZE];

   old_next_read=buf->next_read;
   old_count=eib_get_buf_count(buf);
   for (i=0;i<old_count;i++)
   {  len=eib_read_buf(buf,&text[0]);
      printk(KERN_WARNING "eibdrv: len = %i  ==> ",len);
      if (len>0)
      {  for (k=0;k<len;k++)
            printk("%i ",(int)text[k]);
      }
      else
         printk("empty");
      printk("\n");
   }
   buf->next_read=old_next_read;
   buf->count=old_count;
}


/*  --------------------------------------------------------------------------- */


Eib_wpbuf *eib_get_wpbuf(unsigned long size)
{  Eib_wpbuf *buf;
   
   buf=(Eib_wpbuf *)kmalloc(sizeof(Eib_wpbuf)+size*sizeof(Eib_wproc),GFP_KERNEL);
   if (buf!=NULL)
   {  buf->mem=(Eib_wproc *)(buf+1);
      buf->size=size;
      eib_init_wpbuf(buf);
   }
   return buf;
}

void eib_return_wpbuf(Eib_wpbuf *buf) 
{  kfree(buf); 
}      

int eib_write_wpbuf(Eib_wpbuf *buf,pid_t pid,int flags)
{  unsigned long i;
   Eib_wproc *proc;
   
   if (buf->count==buf->size) return -ENOBUFS;
   
   proc=buf->mem;
   i=0;
   while ((proc->pid!=0)||(proc->flags!=0))
   {  proc++;
      i++;
   }
   
   proc->pid=pid;
   proc->flags=flags;
   
   proc->prev=buf->last;
   proc->next=0;

   if (buf->count==0)
      buf->first=i+1;
   else
      (buf->mem+(buf->last-1))->next=i+1;

      
   buf->count++;
   buf->last=i+1;
   if (buf->next_result==0)
      buf->next_result=i+1;
         
   return 0;
}

int eib_condrm_wpbuf(Eib_wpbuf *buf,pid_t pid)
{  unsigned long i;
   Eib_wproc *proc;
   
   if (buf->count==0) return -ENOPROC;
      
   proc=buf->mem+(buf->first-1);
   i=buf->first;
   
   while ((proc->pid!=pid)&&(proc->next!=0))
   {  i=proc->next;
      proc=buf->mem+(proc->next-1);
   }
   
   if (proc->pid!=pid)                return -ENOPROC;
   if ((proc->flags&FT_PROC_FREE)==0) return -ENORESULT;

   if (buf->first==i) buf->first=proc->next;
   else               (buf->mem+(proc->prev-1))->next=proc->next;

   if (buf->last==i) buf->last=proc->prev;
   else              (buf->mem+(proc->next-1))->prev=proc->prev;
   
   proc->pid=0;
   proc->flags=0;
   buf->count--;
   
   wake_up_interruptible(&wq_fullwproc);
   return proc->result;
}

void eib_result_wpbuf(Eib_wpbuf *buf,int result)
{  Eib_wproc *proc;
   
   if (buf->next_result==0) return;
      
   proc=buf->mem+(buf->next_result-1);
   
   proc->result=result;
   proc->flags|=FT_PROC_FREE;

   buf->next_result=proc->next;
   
   if ((proc->flags&FT_RMPROC_AUTO)!=0)
     eib_condrm_wpbuf(buf,proc->pid);
}

int eib_resize_wpbuf(Eib_wpbuf **buf,unsigned long new_size)
{  unsigned long i,j;
   Eib_wproc *proc,*new_proc;
   Eib_wpbuf *new_buf,*old_buf;
   
   if ((new_buf=eib_get_wpbuf(new_size))==NULL) return -ENOMEM;

   old_buf=(*buf);
   if (new_size>old_buf->size)
   {  memcpy(new_buf,old_buf,sizeof(Eib_wpbuf)+old_buf->size*sizeof(Eib_wproc));
      new_buf->mem=(Eib_wproc *)(new_buf+1);
      new_buf->size=new_size;
   }
   else if (new_size<old_buf->size)
   {  if ((old_buf->count>new_size)||(new_size==0)) 
      {  eib_return_wpbuf(new_buf);
         return -ESIZE;
      }
      proc=old_buf->mem+(old_buf->first-1);
      new_proc=new_buf->mem;
      j=old_buf->first;
      for (i=1;i<=old_buf->count;i++)
      {  new_proc->pid=proc->pid;
         new_proc->flags=proc->flags;
         new_proc->result=proc->result;
         new_proc->next=i+1;
         new_proc->prev=i-1;
         if (old_buf->next_result==j) new_buf->next_result=i;
         new_proc++;
         proc=old_buf->mem+(proc->next-1);
         j=proc->next;
      }
      new_buf->first=1;
      new_buf->last=old_buf->count;
      (new_buf->mem)->prev=0;
      (new_buf->mem+(new_buf->last-1))->next=0;
      new_buf->count=old_buf->count;
      if (old_buf->next_result==0)
         new_buf->next_result=0;
   }
   else 
   {  eib_return_wpbuf(new_buf);
      return 0;
   }
   
   eib_return_wpbuf(*buf);
   *buf=new_buf;
   return 0;
}

unsigned long eib_countinst_wpbuf(Eib_wpbuf *buf,pid_t pid)
{  Eib_wproc *proc;
   unsigned long count=0;
   unsigned long i;
   
   for(i=0,proc=buf->mem;i<buf->size;i++,proc++)
      if ((proc->flags!=0)&&(proc->pid==pid))
         count++;
   return count;
}      

unsigned long eib_get_wpbuf_count(Eib_wpbuf *buf) 
{  return buf->count;  
}

unsigned long eib_get_wpbuf_size (Eib_wpbuf *buf) 
{  return buf->size;    
}

void eib_init_wpbuf(Eib_wpbuf *buf)
{  int i;
   Eib_wproc proc={0,0,0,0};
   
   buf->first=0;
   buf->last=0;
   buf->next_result=0;
   buf->count=0;
   for (i=0;i<buf->size;i++)
      *(buf->mem+i)=proc;
}

void eib_display_wpbuf(Eib_wpbuf *buf)
{  unsigned long i;
   Eib_wproc *proc;
   
   if (buf->count==0) return;
      
   proc=buf->mem+(buf->first-1);
   for (i=0;i<buf->count;i++)
   {  printk(KERN_WARNING "eibdrv: Prozess-ID %li  %x  %i  %li  %li\n",
         (long)proc->pid,proc->flags,proc->result,proc->prev,proc->next);
      proc=buf->mem+(proc->next-1);
   }
}

   
/*  --------------------------------------------------------------------------- */


Eib_rpbuf *eib_get_rpbuf(unsigned long size,unsigned long msgbuf_size)
{  Eib_rpbuf *buf;
   
   if ((buf=(Eib_rpbuf *)kmalloc(sizeof(Eib_rpbuf)+size*(sizeof(Eib_rproc)+
               sizeof(Eib_buf)+msgbuf_size*(EIB_BUF_ELEMENT_SIZE+1)*
               sizeof(unsigned char)),GFP_KERNEL))==NULL)
      return NULL;
   
   buf->mem=(Eib_rproc *)(buf+1);
   buf->size=size;
   buf->msgbuf_size=msgbuf_size;
   eib_init_rpbuf(buf);
   
   return buf;
}

void eib_init_rpbuf(Eib_rpbuf *buf)
{  int i;
   unsigned char *charbuf;
   Eib_rproc *proc;
   Eib_buf *msgbuf;
   
   msgbuf=(Eib_buf *)(buf->mem+buf->size);
   charbuf=(unsigned char *)(msgbuf+buf->size);
  
   buf->count=0;
   for (i=0,proc=buf->mem;i<buf->size;i++,proc++)
   {  proc->filp=NULL;
      proc->flags=0;
      proc->wmode=FT_WRITE_STANDARD;
      proc->msgbuf=msgbuf+i;
      proc->msgbuf->mem=charbuf+i*(buf->msgbuf_size*(EIB_BUF_ELEMENT_SIZE+1));
      proc->msgbuf->size=buf->msgbuf_size;
      eib_init_buf(proc->msgbuf);
   }      
}

void eib_return_rpbuf(Eib_rpbuf *buf)  
{  kfree(buf);  
}

Eib_rproc *eib_find_rproc(Eib_rpbuf *buf,struct file *filp)
{  unsigned long i;
   Eib_rproc *proc;
   
   for (i=0,proc=buf->mem;i<buf->size;i++,proc++)
      if (((proc->flags&FT_PROC_OPEN)==FT_PROC_OPEN)&&(proc->filp==filp))
         return proc;
   return NULL;
}

int eib_ins_rpbuf(Eib_rpbuf *buf,struct file *filp,int flags)
{  Eib_rproc *proc;

   if (buf->count==buf->size) return -ENOBUFS;
   if (eib_find_rproc(buf,filp)!=NULL) return -EDEVISOPEN;
   proc=buf->mem;
   while ((proc->flags&FT_PROC_OPEN)==FT_PROC_OPEN)
      proc++;
   proc->filp=filp;
   proc->flags=flags|FT_PROC_OPEN;
   proc->wmode=FT_WRITE_STANDARD;
   buf->count++;
   eib_init_buf(proc->msgbuf);
   return 0;
}

int eib_rm_rpbuf(Eib_rpbuf *buf,struct file *filp)
{  Eib_rproc *proc;
   
   if (buf->count==0) return -ENOPROC;
   if ((proc=eib_find_rproc(buf,filp))!=NULL)
   {  proc->flags-=FT_PROC_OPEN;
      buf->count--;
      return 0;
   }
   return -ENOPROC;
}

int eib_wmsg_rpbuf(Eib_rpbuf *buf,unsigned char *element,int len)
{  int result=0;
   unsigned long i;
   Eib_rproc *proc;
   
   if (buf->count==0) return 0;
   
   for (i=0,proc=buf->mem;i<buf->size;i++,proc++)
     if ((proc->flags&FT_PROC_OPEN)==FT_PROC_OPEN)
        if (eib_write_buf(proc->msgbuf,element,len)==-ENOBUFS)
           result=-ENOBUFS;
   return result;
}

int eib_rmsg_rpbuf(Eib_rpbuf *buf,struct file *filp,unsigned char *element)
{   Eib_rproc *proc;
    
   if (buf->count==0) return 0;
   if ((proc=eib_find_rproc(buf,filp))!=NULL)
      return eib_read_buf(proc->msgbuf,element);
   return 0;
}

int eib_resize_rpbuf(Eib_rpbuf **buf,unsigned long new_size,unsigned long new_msgbuf_size)
{  unsigned long i;
   Eib_rpbuf *old_buf,*new_buf;
   Eib_rproc *old_proc,*new_proc;
   
   if ((new_buf=eib_get_rpbuf(new_size,new_msgbuf_size))==NULL) return -ENOMEM;

   old_buf=(*buf);
   if ((old_buf->size==new_size)&&(old_buf->msgbuf_size==new_msgbuf_size)) 
   {  eib_return_rpbuf(new_buf);
      return 0;
   }
   if ((new_size==0)||(new_msgbuf_size==0)) 
   {  eib_return_rpbuf(new_buf);
      return -ESIZE;
   }
   if (new_size<old_buf->count) 
   {  eib_return_rpbuf(new_buf);
      return -ESIZE;
   }
   for (i=0,old_proc=old_buf->mem;i<old_buf->size;i++,old_proc++)
      if (((old_proc->flags&FT_PROC_OPEN)==FT_PROC_OPEN)&&(new_msgbuf_size<old_proc->msgbuf->count))
      {  eib_return_rpbuf(new_buf);
         return -ESIZE;
      }
   
   if (old_buf->count>0)
   {  new_buf->count=old_buf->count;
      new_proc=new_buf->mem;
      for (i=0,old_proc=old_buf->mem;i<old_buf->size;i++,old_proc++)
      {  if ((old_proc->flags&FT_PROC_OPEN)==FT_PROC_OPEN)
         {  new_proc->filp=old_proc->filp;
            new_proc->flags=old_proc->flags;
            if (old_proc->msgbuf->count>0)
               eib_copymem_buf(old_proc->msgbuf,new_proc->msgbuf);
            new_proc++;
         }
      }
   }
   
   eib_return_rpbuf(*buf);
   *buf=new_buf;
   return 0;
}

unsigned long eib_get_rpbuf_msgcount(Eib_rpbuf *buf,struct file *filp)
{  Eib_rproc *proc;
   
   if (buf->count==0) return -ENOPROC;
   if ((proc=eib_find_rproc(buf,filp))!=NULL)
      return proc->msgbuf->count;
   return -ENOPROC;
}

unsigned long eib_get_rpbuf_count(Eib_rpbuf *buf)  
{  return buf->count;  
}

unsigned long eib_get_rpbuf_size(Eib_rpbuf *buf)
{  return buf->size;
}

unsigned long eib_get_rpbuf_msgbuf_size(Eib_rpbuf *buf)
{  return buf->msgbuf_size;
}

int eib_set_rpbuf_wmode(Eib_rpbuf *buf,struct file *filp,int wmode)
{  Eib_rproc *proc;
   
   if (buf->count==0) return -ENOPROC;
   if ((proc=eib_find_rproc(buf,filp))!=NULL)
   {  proc->wmode=wmode;
      return 0;
   }
   return -ENOPROC;
}

int eib_get_rpbuf_wmode(Eib_rpbuf *buf,struct file *filp)
{  Eib_rproc *proc;
   
   if (buf->count==0) return -ENOPROC;
   if ((proc=eib_find_rproc(buf,filp))!=NULL)
      return proc->wmode;
   return -ENOPROC;
}


void eib_display_rpbuf(Eib_rpbuf *buf)
{  unsigned long i;
   Eib_rproc *proc;
   
   if (buf->count==0) return;
   for (i=0,proc=buf->mem;i<buf->size;i++,proc++)
      if ((proc->flags&FT_PROC_OPEN)==FT_PROC_OPEN)
      {  printk(KERN_WARNING "eibdrv: Prozess flags %x\n",proc->flags);
         eib_display_buf(proc->msgbuf);
      }
}  

/*  --------------------------------------------------------------------------- */

struct timeval sum_timeval(struct timeval a,struct timeval b)
{  a.tv_sec+=b.tv_sec;
   a.tv_usec+=b.tv_usec;
   if (a.tv_usec>=1000000)
   {  a.tv_sec+=1;
      a.tv_usec-=1000000;
   }
   return a;
}


struct timeval diff_timeval(struct timeval a,struct timeval b)
{  struct timeval t={0,0};
   
   if ((a.tv_sec>b.tv_sec)||((a.tv_sec==b.tv_sec)&&(a.tv_usec>=b.tv_usec)))
   {  a.tv_sec-=b.tv_sec;
      a.tv_usec-=b.tv_usec;
      if (a.tv_usec<0)
      {  a.tv_sec-=1;
         a.tv_usec+=1000000;
      }
      return a;
   }
   else
      return t;
}


unsigned long timeval_to_jiffies(struct timeval t)
{  unsigned long j;
   
   j=t.tv_sec*HZ+t.tv_usec/(1000000/HZ);
   if ((t.tv_usec%(1000000/HZ))!=0)
      j++;
   return j;
}


struct timeval jiffies_to_timeval(unsigned long j)
{  struct timeval t;
   
   t.tv_sec=j/HZ;
   t.tv_usec=(j%HZ)*(1000000/HZ);
   return t;
}

void lock(sv *key)
{  while (key->val==0)
      interruptible_sleep_on(&key->wq);
   key->val--;
}                  
 
void unlock(sv *key)
{  key->val++;
   wake_up_interruptible(&key->wq);
}                   

int check_lock(sv *key)
{  return key->val;
}

void reset_statistics()
{  int i;

   ftstation.stat.outbuf_count=0;
   ftstation.stat.wpbuf_count=0;
   ftstation.stat.rpbuf_count=0;   
   for(i=0;i<4;i++) ftstation.stat.write[i]=0;
   for(i=0;i<3;i++) ftstation.stat.read[i]=0;
   ftstation.stat.total_w1time.tv_sec=0;
   ftstation.stat.total_w1time.tv_usec=0;
   ftstation.stat.overflow=0;
   
   ftstation.rrepeat=-1;
}

void line_idle_delay(void)
{  struct timeval t;
   time_t delay_sec;
   suseconds_t delay_usec;
   
   do_gettimeofday(&t);
   
   if ((delay_sec=t.tv_sec-ftstation.last_seract.tv_sec) > ftstation.line_idle_timeout_sec) 
      return;

   delay_usec=delay_sec*1000000+t.tv_usec-ftstation.last_seract.tv_usec;
   
   if (delay_usec >= ftstation.line_idle_timeout_usec) return;
   udelay(ftstation.line_idle_timeout_usec-delay_usec);
}

int immediate_read(unsigned char *buf) 
{
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)      
   if  (ftstation.port.filp->f_op->read(ftstation.port.filp,buf,1,
                      &ftstation.port.filp->f_pos)==0)
#else
   if (ftstation.port.filp->f_op->read(ftstation.port.inodp,ftstation.port.filp,buf,1)==0)
#endif
      return -EAGAIN;
   
   return 1;
}      

int timeout_read(unsigned char *buf,int count)
{  int n_read=0;
   int result;
   struct termios help_termios;

   ftstation.port.tty->termios->c_cc[VTIME]=1;
   ftstation.port.tty->termios->c_cc[VMIN]=0;
   ftstation.port.tty->driver.set_termios(ftstation.port.tty,&help_termios);
   ftstation.port.tty->ldisc.set_termios(ftstation.port.tty,&help_termios);

   while (n_read<count)
   {  
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)      
      if  ((result=ftstation.port.filp->f_op->read(ftstation.port.filp,buf+n_read,
                  count-n_read,&ftstation.port.filp->f_pos))==0)
#else
      if ((result=ftstation.port.filp->f_op->read(ftstation.port.inodp,ftstation.port.filp,&buf[n_read],
                  count-n_read))==0)            
#endif
      {  ftstation.port.tty->termios->c_cc[VTIME]=0;
         ftstation.port.tty->termios->c_cc[VMIN]=0;
         ftstation.port.tty->driver.set_termios(ftstation.port.tty,&help_termios);
         ftstation.port.tty->ldisc.set_termios(ftstation.port.tty,&help_termios);
         return n_read;
      }
      n_read+=result;
   }

   ftstation.port.tty->termios->c_cc[VTIME]=0;
   ftstation.port.tty->termios->c_cc[VMIN]=0;
   ftstation.port.tty->driver.set_termios(ftstation.port.tty,&help_termios);
   ftstation.port.tty->ldisc.set_termios(ftstation.port.tty,&help_termios);

   return n_read;
}

int repeat_write(unsigned char *msg,int msg_len,int wait_state)
{  unsigned char ack_msg[4];
   int stat_write=0;
   struct timeval wait_start,wait_actual;
   unsigned long wait_jiffies;
   
   ftstation.wrepeat=0;
   ftstation.state=wait_state;
   if (wait_state==FT_STATE_WAIT_FOR_UDATA_ACK)
      stat_write=1;
      
   fs=get_fs();
   set_fs(KERNEL_DS);
         
   do
   {  ftstation.wrepeat++;
#ifdef DEBUG      
      printk(KERN_WARNING "eibdrv (%i): Schreibversuch %i\n",c++,ftstation.wrepeat);
#endif      
      line_idle_delay();
      do_gettimeofday(&ftstation.start_write);
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)      
      ftstation.port.filp->f_op->write(ftstation.port.filp,msg,msg_len,&ftstation.port.filp->f_pos);
#else
      ftstation.port.filp->f_op->write(ftstation.port.inodp,ftstation.port.filp,msg,msg_len);
#endif            
      do_gettimeofday(&ftstation.last_seract);

      switch(mode)
      {  case FT_SERVERMODE:
                 unlock(&sv_ser_read);
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
                 interruptible_sleep_on_timeout(&wq_wait_for_ack,ftstation.exchange_timeout);      
#else
                 current->timeout=jiffies+ftstation.exchange_timeout;
                 interruptible_sleep_on(&wq_wait_for_ack);
#endif         
                 lock(&sv_ser_read);
                 break;
         case FT_STANDARDMODE:
                 wait_start=ftstation.last_seract;
                 wait_actual=wait_start;
                 while (((wait_jiffies=timeval_to_jiffies(diff_timeval(ftstation.exchange_time,
                           diff_timeval(wait_actual,wait_start))))>0)&&(ftstation.state!=FT_STATE_READY))
                 {
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
                    interruptible_sleep_on_timeout(&ftstation.port.tty->read_wait,wait_jiffies);
#else
                    current->timeout=jiffies+wait_jiffies;
                    interruptible_sleep_on(&ftstation.port.tty->read_wait);
#endif         
                    do_gettimeofday(&wait_actual);
                    while ((immediate_read(&ack_msg[0])==1)&&(ftstation.state!=FT_STATE_READY))
                    {  switch(ack_msg[0])
                       {  case FT_FIXED_LENGTH_MSG:
                                  if (timeout_read(&ack_msg[0],3)!=3) break;
                                  if ((ack_msg[0]!=ack_msg[1])||(ack_msg[2]!=0x16)) break;
                                  if ((ack_msg[0]&0xDF)==(0x0B|((1-FT_DIR)<<7)))
                                  {  
#ifdef DEBUG                                     
                                     printk(KERN_WARNING "eibdrv (%i): RESP_STATUS  ==>  erhalten\n",c++);
#endif
                                     if (ftstation.state==FT_STATE_WAIT_FOR_RESP_STATUS)
                                        ftstation.state=FT_STATE_READY;
                                  }
                                  break;
                          case FT_ACK:
                                  do_gettimeofday(&ftstation.end_write);
                                  if (ftstation.wrepeat==1)
                                     ftstation.stat.total_w1time=sum_timeval(ftstation.stat.total_w1time,
                                            diff_timeval(ftstation.end_write,ftstation.start_write));
#ifdef DEBUG                                            
                                  printk(KERN_WARNING "eibdrv (%i): ACK  ==>  erhalten!\n",c++);
#endif                                  
                                  if ((ftstation.state==FT_STATE_WAIT_FOR_UDATA_ACK)||
                                      (ftstation.state==FT_STATE_WAIT_FOR_RESET_ACK))
                                     ftstation.state=FT_STATE_READY;
                                  break;
                       }
                    }
                    break;
                 }
      }
   }
   while ((ftstation.state==wait_state)&&(ftstation.wrepeat<FT_REPEAT_LIMIT));

   set_fs(fs);         

   if (ftstation.state==wait_state)
   {  ftstation.state=FT_STATE_READY;
      if (stat_write==1)
      {  if (ftstation.stat.write[0]<ULONG_MAX)
            ftstation.stat.write[0]++;
         else
            ftstation.stat.overflow=1;
      }
      return 0;
   }
   else
   {  if (stat_write==1)
      {  if (ftstation.stat.write[ftstation.wrepeat]<ULONG_MAX)
            ftstation.stat.write[ftstation.wrepeat]++;
         else
            ftstation.stat.overflow=1;
      }
      return msg_len;
   }
}

void ser_write(void)
{  int udata_len;
   int msg_len;
   int checksum;
   int i,result;
   unsigned char msg[EIB_BUF_ELEMENT_SIZE+7];

   do
   {  interruptible_sleep_on(&wq_ser_write);
//    interruptible_sleep_on_timeout(&sleep_queue,20*HZ);
      lock(&sv_ser_write);
      lock(&sv_ser_read);
      
      do
      {  if ((udata_len=eib_read_buf(ftstation.outbuf,&msg[5]))==0)
         {  unlock(&sv_ser_read);
            unlock(&sv_ser_write);
            break;
         }

         msg_len=udata_len+7;
         msg[0]=msg[3]=0x68;
         msg[1]=msg[2]=(unsigned char)(udata_len+1);
         msg[4]=(FT_DIR<<7) | 0x40 | (ftstation.fcb<<5) | (FT_FCB_VALID<<4) | 0x03;
         checksum=msg[4];
#ifdef DEBUG         
         printk("<1>Gesendetes fcb=%i\n",ftstation.fcb);
#endif DEBUG         
         for (i=0;i<udata_len;i++)
         {   checksum+=msg[i+5];
         }
         msg[5+udata_len]=(unsigned char) checksum % 256;
         msg[6+udata_len]=0x16;
       
         if ((result=repeat_write(&msg[0],msg_len,FT_STATE_WAIT_FOR_UDATA_ACK))==msg_len)
            result=udata_len;
         if (result>=1)
            ftstation.fcb=1-ftstation.fcb;
         eib_result_wpbuf(ftstation.wpbuf,result);
         wake_up_interruptible(&wq_wproc);
      }
      while(1);
   }
   while (ftstation.state!=FT_STATE_SHUTDOWN);
}

int ser_reset(void)
{  int result=0;
   unsigned char msg[4];
   
   msg[0]=0x10;
   msg[1]=msg[2]=(0x40 | (FT_DIR << 7));
   msg[3]=0x16;

   lock(&sv_ser_write);
   lock(&sv_ser_read);
   
   if (repeat_write(&msg[0],4,FT_STATE_WAIT_FOR_RESET_ACK)==0)
      result=-ENORESET;

   unlock(&sv_ser_read);
   unlock(&sv_ser_write);      
   
   return result;
}

int ser_status(void)
{  int result=0;
   unsigned char msg[4];
   
   msg[0]=0x10;
   msg[1]=msg[2]=(0x49 | (FT_DIR << 7));
   msg[3]=0x16;

   lock(&sv_ser_write);
   lock(&sv_ser_read);

   if (repeat_write(&msg[0],4,FT_STATE_WAIT_FOR_RESP_STATUS)==0)
      result=-ENORESPSTATUS;

   unlock(&sv_ser_read);
   unlock(&sv_ser_write);      
      
   return result;
}

void empty_ttyS() 
{  unsigned char msg;
   struct termios help_termios;
   
   ftstation.port.tty->termios->c_cc[VTIME]=0;
   ftstation.port.tty->termios->c_cc[VMIN]=0;
          
   ftstation.port.tty->driver.set_termios(ftstation.port.tty,&help_termios);
   ftstation.port.tty->ldisc.set_termios(ftstation.port.tty,&help_termios);

   while (ftstation.port.filp->f_op->read(ftstation.port.filp,&msg,1,&ftstation.port.filp->f_pos)!=0)
      ;
   
   ftstation.port.tty->termios->c_cc[VTIME]=1;
   ftstation.port.tty->termios->c_cc[VMIN]=0;
          
   ftstation.port.tty->driver.set_termios(ftstation.port.tty,&help_termios);
   ftstation.port.tty->ldisc.set_termios(ftstation.port.tty,&help_termios);
}      
   
void ser_read(void)
{  int checksum;
   int msg_fcb;
   int i,k=0;
   int udata_len;
   unsigned char msg[EIB_BUF_ELEMENT_SIZE+7];
   
   do
   {  k++;
      interruptible_sleep_on(&ftstation.port.tty->read_wait);
      if (ftstation.state==FT_STATE_SHUTDOWN)
         break;
      fs=get_fs();
      set_fs(KERNEL_DS);
       
//    if (timeout_read(&msg[0],1)==1)
      while (immediate_read(&msg[0])==1)      
      {  do_gettimeofday(&ftstation.end_write);         
         
         switch(msg[0])
         {  case FT_VAR_LENGTH_MSG: // SEND_UDATA received
                    if (timeout_read(&msg[0],4)!=4) break;
                    if ((msg[0]!=msg[1]) || (msg[0]>24) || (msg[2]!=0x68) ||
                       ((msg[3]&0xDF)!=(unsigned char)(((1-FT_DIR)<<7) | 0x40 | (FT_FCB_VALID<<4) | 0x03)))
                       break;

                    msg_fcb=(int)((msg[3] & 0x20)>>5);
                    checksum=msg[3];
                    udata_len=msg[0]-1;

                    if (timeout_read(&msg[0],udata_len+2)!=udata_len+2) break;
                    for (i=0;i<udata_len;i++)
                       checksum+=msg[i];
                    if ((msg[udata_len]!=(unsigned char)(checksum%256)) || (msg[udata_len+1]!=0x16))
                       break;
                    if (msg_fcb==(1-ftstation.rec_fcb))   // msg not already received
                    {  
#ifdef DEBUG                       
                       printk(KERN_WARNING "eibdrv (%i): SEND_UDATA  ==>  neue Nachricht\n",c++);
                       printk("<1>fcb erwartet = %i     erhalten = %i\n",(1-ftstation.rec_fcb),msg_fcb); 
#endif                       
                       ftstation.rec_fcb=1-ftstation.rec_fcb;
                       ftstation.rrepeat=0;
                       eib_wmsg_rpbuf(ftstation.rpbuf,&msg[0],udata_len);
                       wake_up_interruptible(&wq_rproc);
                       if (ftstation.pasync!=NULL)
                          kill_fasync(ftstation.pasync,SIGIO);
                    }
                    else 
                    {  
#ifdef DEBUG                       
                       printk(KERN_WARNING "eibdrv (%i): SEND_UDATA  ==>  bereits erhalten\n",c++);
                       printk("<1>fcb erwartet = %i     erhalten = %i\n",(1-ftstation.rec_fcb),msg_fcb); 
#endif                       
                       if (ftstation.rrepeat!=-1)
                          ftstation.rrepeat++;
                    }
                    
                    if (ftstation.rrepeat<3)
                    {  if (ftstation.stat.read[ftstation.rrepeat]<ULONG_MAX)
                          ftstation.stat.read[ftstation.rrepeat]++;
                       else
                          ftstation.stat.overflow=1;
                       if (ftstation.rrepeat>0)
                          ftstation.stat.read[ftstation.rrepeat-1]--;  
                    }
                    line_idle_delay();
                    msg[0]=FT_ACK;
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)                    
                    ftstation.port.filp->f_op->write(ftstation.port.filp,&msg[0],1,&ftstation.port.filp->f_pos);
#else
                    ftstation.port.filp->f_op->write(ftstation.port.inodp,ftstation.port.filp,&msg[0],1);
#endif                    
                    do_gettimeofday(&ftstation.last_seract);
                    break;
                    
            case FT_FIXED_LENGTH_MSG:
                    if (timeout_read(&msg[0],3)!=3) break;
                    if ((msg[0]!=msg[1]) || (msg[2]!=0x16)) break;
                    switch(msg[0] & 0xDF)
                    {  case (0x40 | ((1-FT_DIR)<<7)): // SEND_RESET
#ifdef DEBUG                       
                               printk(KERN_WARNING "eibdrv (%i): SEND_RESET  ==>  erhalten!\n",c++);
#endif                               
                               ftstation.fcb=1;
                               ftstation.rec_fcb=0;
                               ftstation.wrepeat=FT_REPEAT_LIMIT;
                               line_idle_delay();
                               msg[0]=FT_ACK;
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)                               
                               ftstation.port.filp->f_op->write(ftstation.port.filp,&msg[0],1,
                                  &ftstation.port.filp->f_pos);
#else
                               ftstation.port.filp->f_op->write(ftstation.port.inodp,ftstation.port.filp,&msg[0],1);
#endif                                                                    
                               do_gettimeofday(&ftstation.last_seract);
                               break;
                       case (0x49 | ((1-FT_DIR)<<7)): // REQ_STATUS
#ifdef DEBUG                       
                               printk(KERN_WARNING "eibdrv (%i): REQ_STATUS  ==>  erhalten!\n",c++);
#endif DEBUG                               
                               line_idle_delay();
                               msg[0]=0x10;
                               msg[1]=msg[2]=(0x0B | (FT_DIR<<7));
                               msg[3]=0x16;
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)                               
                               ftstation.port.filp->f_op->write(ftstation.port.filp,&msg[0],4,
                                  &ftstation.port.filp->f_pos);
#else
                               ftstation.port.filp->f_op->write(ftstation.port.inodp,ftstation.port.filp,&msg[0],4);
#endif                                                                    
                               do_gettimeofday(&ftstation.last_seract);
                               break;
                       case (0x0B | ((1-FT_DIR)<<7)): // RESP_STATUS
#ifdef DEBUG                        
                               printk(KERN_WARNING "eibdrv (%i): RESP_STATUS  ==>  erhalten!\n",c++);
#endif                               
                               if (ftstation.state==FT_STATE_WAIT_FOR_RESP_STATUS)
                                  ftstation.state=FT_STATE_READY;
                               break;
                    }
                    break;
                          
            case FT_ACK:            // ACK received 
                    if (ftstation.wrepeat==1)
                       ftstation.stat.total_w1time=sum_timeval(ftstation.stat.total_w1time,
                                 diff_timeval(ftstation.end_write,ftstation.start_write));
#ifdef DEBUG                                 
                    printk(KERN_WARNING "eibdrv (%i): ACK  ==>  erhalten\n",c++);
#endif                    
                    if ((ftstation.state==FT_STATE_WAIT_FOR_UDATA_ACK) ||
                        (ftstation.state==FT_STATE_WAIT_FOR_RESET_ACK))
                       ftstation.state=FT_STATE_READY;
                    break;
         }  
//       empty_ttyS();
         if ((ftstation.state==FT_STATE_WAIT_FOR_UDATA_ACK) ||
             (ftstation.state==FT_STATE_WAIT_FOR_RESET_ACK) ||
             (ftstation.state==FT_STATE_WAIT_FOR_RESP_STATUS))
            wake_up_interruptible(&wq_wait_for_ack);
      }
      set_fs(fs);
   }
   while(1);
}


int open_ftstation(void)
{  
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)   
   struct dentry *ser_dentry=NULL;
#endif
   
   struct inode *ser_inode=NULL;
   struct async_struct ser_async;
   struct termios help_termios;

   fs=get_fs();
   set_fs(KERNEL_DS);

#ifdef __SMP__
   lock_kernel();
#endif

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)                                  
   ftstation.port.filp=filp_open(port,O_RDWR|O_NOCTTY|O_EXCL,0600);
   if (IS_ERR(ftstation.port.filp)) 
   {  printk(KERN_ERR "eibdrv (%i): cannot open serial port %s\n",c++,port);
      goto fail_open_serport;
   }
   ser_dentry=ftstation.port.filp->f_dentry;
   ftstation.port.inodp=ser_inode=ser_dentry->d_inode;
#else
   ftstation.port.filp=open_fp(port,O_RDWR|O_NOCTTY|O_EXCL,0600);
   if (ftstation.port.filp==NULL)
   {  printk(KERN_ERR "eibdrv (%i): cannot open serial port %s\n",c++,port);
      goto fail_open_serport;
   }
   ftstation.port.inodp=ser_inode=ftstation.port.filp->f_inode;   
#endif
    
   if (ftstation.port.filp->f_op->ioctl(ser_inode,ftstation.port.filp,
                TIOCSERGSTRUCT,(unsigned long)&ser_async)<0)
   {  printk(KERN_ERR "eibdrv (%i): cannot get async_struct\n",c++);
      goto fail_get_async;
   }
     
   ftstation.port.tty=ser_async.tty;
   ftstation.port.old_termios=*(ftstation.port.tty->termios);      
      
   ftstation.port.tty->termios->c_iflag=IGNPAR |         // ignore bytes with parity or framing errors
                                        INPCK;           // enable input parity checking 
   ftstation.port.tty->termios->c_oflag=0x00;
   ftstation.port.tty->termios->c_cflag=symb_baudrate |  // baudrate
                                        CRTSCTS |        // hardware flow control
                                        CS8 |            // character size 8 bit
                                        PARENB |         // parity bit 
                                        CLOCAL |         // ignore modem control lines
                                        CREAD;           // enable receiver
   ftstation.port.tty->termios->c_lflag=0x00;
   ftstation.port.tty->termios->c_cc[VTIME]=0;
   ftstation.port.tty->termios->c_cc[VMIN]=0;
          
   ftstation.port.tty->driver.set_termios(ftstation.port.tty,&help_termios);
   ftstation.port.tty->ldisc.set_termios(ftstation.port.tty,&help_termios);

#ifdef __SMP__
   unlock_kernel();
#endif
   
   if (mode==FT_SERVERMODE)
   {  if ((ftstation.wpbuf=eib_get_wpbuf(wpbuf_size))==NULL)
      {  printk(KERN_ERR "eibdrv (%i): cannot allocate memory for list of write processes\n",c++);
         goto fail_wpbuf;
      }
      
      if ((ftstation.rpbuf=eib_get_rpbuf(rpbuf_size,rpbuf_msgbuf_size))==NULL)
      {  printk(KERN_ERR "eibdrv (%i): cannot allocate memory for list of read processes\n",c++);
         goto fail_rpbuf;
      }
       
      if ((ftstation.outbuf=eib_get_buf(outbuf_size))==NULL)
      {  printk(KERN_ERR "eibdrv (%i): cannot allocate memory for message buffer\n",c++);
         goto fail_outbuf;
      }
   }
   
   set_fs(fs);
      
   return 0;      

   fail_outbuf:       eib_return_rpbuf(ftstation.rpbuf);
   fail_rpbuf:        eib_return_wpbuf(ftstation.wpbuf);
   fail_wpbuf:        *(ftstation.port.tty->termios)=ftstation.port.old_termios;
                      ftstation.port.tty->driver.set_termios(ftstation.port.tty,&help_termios);
                      ftstation.port.tty->ldisc.set_termios(ftstation.port.tty,&help_termios);
   fail_get_async:    
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0) 
                      filp_close(ftstation.port.filp,NULL);
#else
                      close_fp(ftstation.port.filp);
#endif

   fail_open_serport: 
#ifdef __SMP__
                      unlock_kernel();
#endif
                      set_fs(fs);
                      return -ESHUTDOWN;
}

int close_ftstation(void)
{  struct termios help_termios;
   
   if (mode==FT_SERVERMODE)
   {  eib_return_buf(ftstation.outbuf);
      eib_return_rpbuf(ftstation.rpbuf);
      eib_return_wpbuf(ftstation.wpbuf);
   }
   
   fs=get_fs();
   set_fs(KERNEL_DS);

#ifdef __SMP__
   lock_kernel();
#endif
         
   *(ftstation.port.tty->termios)=ftstation.port.old_termios;      
      
   ftstation.port.tty->driver.set_termios(ftstation.port.tty,&help_termios);
   ftstation.port.tty->ldisc.set_termios(ftstation.port.tty,&help_termios);

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0) 
   filp_close(ftstation.port.filp,NULL);
   if (IS_ERR(ftstation.port.filp)) 
   {  printk(KERN_ERR "eibdrv (%i): cannot close serial port %s\n",c++,port);
      goto fail_close_serport;
   }
#else
   if (close_fp(ftstation.port.filp)!=0)
   {  printk(KERN_ERR "eibdrv (%i): cannot close serial port %s\n",c++,port);
      goto fail_close_serport;
   }
#endif      

#ifdef __SMP__
   unlock_kernel();
#endif
   
   set_fs(fs);
   return 0;
      
   fail_close_serport: 
   
#ifdef __SMP__
                       unlock_kernel();
#endif
      		       set_fs(fs);
                       return -ESHUTDOWN;
}


int eibdrv_open(struct inode *eib_inode,struct file *eib_filp)
{  int blockmode;
   int result;

   switch(mode)
   {  case FT_SERVERMODE:
              switch(ftstation.state)
              {  case FT_STATE_INIT:
                         if (ftstation.servproc==2) return -EBUSY;
                         ftstation.servproc++;
                         return 0;
                 case FT_STATE_SHUTDOWN:
                         return -EBUSY;
                 default:
                         if (!MOD_IN_USE)          // first open
                         {  if ((eib_filp->f_flags & O_EXCL)!=0) 
                               ftstation.f_exclusive=FT_EXCLUSIVE_RW;
                            else 
                               ftstation.f_exclusive=FT_SHARED_RW;
                         }
                         else
                         {  if (((eib_filp->f_flags & O_EXCL)!=0) || (ftstation.f_exclusive==FT_EXCLUSIVE_RW))
                               return -EBUSY;
                         }
                         
                         if (!(eib_filp->f_flags&O_NONBLOCK))  blockmode=FT_PROC_BLOCKED;
                         else                                  blockmode=FT_PROC_NONBLOCKED;

                         result=eib_ins_rpbuf(ftstation.rpbuf,eib_filp,blockmode);
                         if (!result) MOD_INC_USE_COUNT;
                         return result;
              }
      case FT_STANDARDMODE:
              if (!MOD_IN_USE)          // first open
              {  if ((eib_filp->f_flags & O_EXCL)!=0) 
                    ftstation.f_exclusive=FT_EXCLUSIVE_RW;
                 else 
                    ftstation.f_exclusive=FT_SHARED_RW;
              }
              else
              {  if (((eib_filp->f_flags & O_EXCL)!=0) || (ftstation.f_exclusive==FT_EXCLUSIVE_RW))
                    return -EBUSY;
              }
                         
              MOD_INC_USE_COUNT;
              return 0;
   } 
   return 0;
}

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
int eibdrv_release(struct inode *eib_inodp,struct file *eib_filp)
#else
void eibdrv_release(struct inode *eib_inodp,struct file *eib_filp)
#endif
{  switch(mode)
   {  case FT_SERVERMODE:
              switch(ftstation.state)
              {  case FT_STATE_SHUTDOWN:
                         ftstation.servproc--;
                         wake_up_interruptible(&wq_shutdown);
                         break;
                 default:
                         MOD_DEC_USE_COUNT;
                         eib_rm_rpbuf(ftstation.rpbuf,eib_filp); 
                         break;
              }
              break;
      case FT_STANDARDMODE:
              MOD_DEC_USE_COUNT;
              break;
   }
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
   return 0;
#endif
}

    

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
ssize_t eibdrv_write(struct file *eib_filp,const char *buf,size_t len,loff_t *pos)
#else
read_write_t eibdrv_write(struct inode *eib_inodp, struct file *eib_filp,const char *buf,int len)
#endif
{  unsigned char element[EIB_BUF_ELEMENT_SIZE];
   int blockmode;
   int result;
   int copylen;
   int wmode,rmmode;

   if (len>23) copylen=23; else copylen=len;

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
   if (copy_from_user(&element[0],buf,copylen)!=0)
      return -EFAULT;
#else
   memcpy_fromfs(&element[0],buf,copylen);
#endif
      
   if (!(eib_filp->f_flags & O_NONBLOCK)) blockmode=FT_PROC_BLOCKED;
   else                                   blockmode=FT_PROC_NONBLOCKED;
      
   wmode=eib_get_rpbuf_wmode(ftstation.rpbuf,eib_filp);
   rmmode=wmode+1;

   while((ftstation.wpbuf->count==wpbuf_size)||(ftstation.outbuf->count==outbuf_size))
   {  if (blockmode==FT_PROC_NONBLOCKED)
         return -EAGAIN;
      else
         interruptible_sleep_on(&wq_fullwproc);
   }
   
   eib_write_buf(ftstation.outbuf,&element[0],copylen);
   eib_write_wpbuf(ftstation.wpbuf,current->pid,rmmode);
   wake_up_interruptible(&wq_ser_write);
   
   if (wmode==FT_WRITE_STANDARD) return copylen;

   do
   {  interruptible_sleep_on(&wq_wproc);
      result=eib_condrm_wpbuf(ftstation.wpbuf,current->pid);
   }
   while (result==-ENORESULT);
   
   return result;
}

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
ssize_t eibdrv_read(struct file *eib_filp,char *buf,size_t len,loff_t *pos)
#else
read_write_t eibdrv_read(struct inode *eib_inodp,struct file *eib_filp,char *buf,int len)
#endif
{  unsigned char element[EIB_BUF_ELEMENT_SIZE];
   int blockmode;
   int copylen;

   if (!(eib_filp->f_flags & O_NONBLOCK)) blockmode=FT_PROC_BLOCKED;
   else                                   blockmode=FT_PROC_NONBLOCKED;

   do
   {  copylen=eib_rmsg_rpbuf(ftstation.rpbuf,eib_filp,&element[0]);
      if (copylen>0)
      {  if (copylen>len) copylen=len;
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)         
         if (copy_to_user(buf,&element[0],copylen)!=0)
            return -EFAULT;
#else
         memcpy_tofs(buf,&element[0],copylen);
#endif
         
         return copylen;
      }

      if (blockmode==FT_PROC_NONBLOCKED) return -EAGAIN;

      interruptible_sleep_on(&wq_rproc);
   }     
   while(1);
}

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
int eibdrv_fasync(int eib_fd,struct file *eib_filp,int flag)
{  return fasync_helper(eib_fd,eib_filp,flag,&(ftstation.pasync));
}
#else
int eibdrv_fasync(struct inode *eib_inodp,struct file *eib_filp,int flag)
{  return fasync_helper(eib_inodp,eib_filp,flag,&(ftstation.pasync));
}
#endif

int eibdrv_ioctl(struct inode *eib_inode,struct file *eib_filp,
                   unsigned int cmd,unsigned long arg)
{  int result;
   Eib_statistics stat;
/* 
   if (current->euid!=0)   // cmd only available to root user
   {  switch(cmd)
      {  case FT_SET_OUTBUF_SIZE:
         case FT_SET_WPBUF_SIZE:
         case FT_SET_RPBUF_SIZE:
         case FT_SET_RPBUF_MSGSIZE:
         case FT_GET_BAUDRATE:
         case FT_GET_PORT_LENGTH:
         case FT_GET_PORT:
         case FT_ISSERVPROC:
         case FT_RESET:
         case FT_RESET_STAT:
         case FT_GET_STAT:
         case FT_GETRESET_STAT: 
                 return -EACCES;
      }
   }
  */       

   switch(cmd)
   {  case FT_GET_MODE :
              return mode;
      case FT_GET_BAUDRATE :
              return symb_baudrate;
      case FT_GET_PORT_LENGTH :
              return strlen(port);
      case FT_GET_PORT:
      
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)               
              if (copy_to_user((char *)arg,port,strlen(port)+1)!=0)
                 return -EFAULT;
#else
              memcpy_tofs((void *)arg,port,strlen(port)+1);            
#endif
              return strlen(port);
              
      case FT_RESET :
              return ser_reset();
      case FT_RESP_STATUS :
              return ser_status();      
              
      case FT_RESET_STAT :
              reset_statistics(&ftstation.stat);
              return 0;
      case FT_GET_STAT :
      case FT_GETRESET_STAT :
              if (ftstation.stat.overflow>0) 
                 return -EOVERFLOW;

              stat=ftstation.stat;    // copy_to_user may block!                 
              if (mode==FT_SERVERMODE)
              {  stat.outbuf_count=eib_get_buf_count(ftstation.outbuf);
                 stat.wpbuf_count=eib_get_wpbuf_count(ftstation.wpbuf);
                 stat.rpbuf_count=eib_get_rpbuf_count(ftstation.rpbuf);
              }

              if (cmd==FT_GETRESET_STAT) 
                 reset_statistics(&ftstation.stat);
                 
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
              if (copy_to_user((Eib_statistics *)arg,&stat,sizeof(Eib_statistics))!=0)
                 return -EFAULT;
#else
              memcpy_to_fs((void *arg),&stat,sizeof(Eib_statistics));
#endif
              return 0;                                          
   }
     
   switch(mode)
   {  case FT_SERVERMODE:
              switch(cmd)
              {  case FT_GET_OUTBUF_SIZE :
                         result=eib_get_buf_size(ftstation.outbuf);
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
                         if (copy_to_user((unsigned long *)arg,&result,sizeof(unsigned long))!=0)
                            return -EFAULT;
#else
                         memcpy_to_fs((void *arg),&result,sizeof(unsigned long));
#endif
                         return 0;
                 case FT_GET_WPBUF_SIZE :    
                         result=eib_get_wpbuf_size(ftstation.wpbuf);
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
                         if (copy_to_user((unsigned long *)arg,&result,sizeof(unsigned long))!=0)
                            return -EFAULT;
#else
                         memcpy_to_fs((void *arg),&result,sizeof(unsigned long));
#endif
                         return 0;
                 case FT_GET_RPBUF_SIZE :    
                         result=eib_get_rpbuf_size(ftstation.rpbuf);
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
                         if (copy_to_user((unsigned long *)arg,&result,sizeof(unsigned long))!=0)
                            return -EFAULT;
#else
                         memcpy_to_fs((void *arg),&result,sizeof(unsigned long));
#endif
                         return 0;
                 case FT_GET_RPBUF_MSGSIZE :    
                         result=eib_get_rpbuf_msgbuf_size(ftstation.rpbuf);
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
                         if (copy_to_user((unsigned long *)arg,&result,sizeof(unsigned long))!=0)
                            return -EFAULT;
#else
                         memcpy_to_fs((void *arg),&result,sizeof(unsigned long));
#endif
                         return 0;
                         
                 case FT_SET_OUTBUF_SIZE :
                         result=eib_resize_buf(&ftstation.outbuf,arg);
                         return result;
                 case FT_SET_WPBUF_SIZE :
                         result=eib_resize_wpbuf(&ftstation.wpbuf,arg);
                         return result;
                 case FT_SET_RPBUF_SIZE :
                         result=eib_resize_rpbuf(&ftstation.rpbuf,arg,ftstation.rpbuf->msgbuf_size);
                         return result;
                 case FT_SET_RPBUF_MSGSIZE :
                         result=eib_resize_rpbuf(&ftstation.rpbuf,ftstation.rpbuf->size,arg);
                         return result;

                 case FT_SET_WRITE_MODE :
                         if ((arg!=FT_WRITE_STANDARD)&&(arg!=FT_WRITE_ACKNOWLEDGED))
                            return -EINVAL;
                         return eib_set_rpbuf_wmode(ftstation.rpbuf,eib_filp,arg);
                 case FT_GET_WRITE_MODE :
                         return eib_get_rpbuf_wmode(ftstation.rpbuf,eib_filp);

                 case FT_ISSERVPROC :
                         if (ftstation.state==FT_STATE_INIT)
                         {  ftstation.servfunc++;
                            if (ftstation.servfunc==1) 
                               ser_write();
                            else
                            {  ftstation.state=FT_STATE_READY;
                               ser_read();
                            }
                 
                            ftstation.servfunc--;
                            return 0;
                         }
                         else
                            return -EINVAL;

              }
              break;

      case FT_STANDARDMODE:
              break;
   }
   
   return -EINVAL;
   
}


              
                 
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
int eibdrv_fsync(struct file *eib_filp,struct dentry *eib_dentry)
#else
int eibdrv_fsync(struct inode *eib_inodp,struct file *eib_filp)
#endif
{  do
   {  if (eib_countinst_wpbuf(ftstation.wpbuf,current->pid)==0)
         return 0;
      interruptible_sleep_on(&wq_fullwproc);
   }
   while (1);      
}

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
loff_t eibdrv_llseek(struct file *eib_filp,loff_t off, int whence)
#else
int eibdrv_lseek(struct inode *eib_inodp,struct file *eib_filp,off_t off, int whence)
#endif
{  return -ESPIPE;
}

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,23)
unsigned int eibdrv_poll(struct file *eib_filp,struct poll_table_struct *eib_polltable)
{  unsigned int mask=0;
   
   poll_wait(eib_filp,&wq_rproc,eib_polltable);
   poll_wait(eib_filp,&wq_fullwproc,eib_polltable);
   
   if ((ftstation.outbuf->count < ftstation.outbuf->size) && (ftstation.wpbuf->count < ftstation.wpbuf->size))
      mask |= POLLOUT | POLLWRNORM;

   if (eib_get_rpbuf_msgcount(ftstation.rpbuf,eib_filp)>0)
      mask |= POLLIN | POLLRDNORM;
      
   return mask;
}      
#else
int eibdrv_select(struct inode *eib_inodp,struct file *eib_filp,int mode,select_table *eib_selecttable)
{  switch(mode)
   {  case SEL_IN:
              if (eib_get_rpbuf_msgcount(ftstation.rpbuf,eib_filp)>0) return 1;
              select_wait(&wq_rproc,eib_selecttable);
              break;
      case SEL_OUT:
              if ((ftstation.outbuf->count < ftstation.outbuf->size) && 
                  (ftstation.wpbuf->count < ftstation.wpbuf->size)) return 1;
              select_wait(&wq_fullwproc,eib_selecttable);
              break;
      case SEL_EX:
              break;
   }
   return 0;
}
#endif

struct file_operations eibdrv_fops =
   { 
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
     eibdrv_llseek,        // llseek
#else
     eibdrv_lseek,         // lseek
#endif

     eibdrv_read,          // read
     eibdrv_write,         // write
     NULL,                 // readdir

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,23)
     eibdrv_poll,          // poll
#else
     eibdrv_select,        // select
#endif   

     eibdrv_ioctl,         // ioctl
     NULL,                 // mmap
     eibdrv_open,          // open

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
     NULL,                   // flush
#endif   

     eibdrv_release,       // release
     eibdrv_fsync,         // fsync
     eibdrv_fasync,        // fasync
     NULL,                 // check_media_change
     NULL,                 // revalidate
     
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
     NULL,                 // lock
#endif     
   };

/*  --------------------------------------------------------------------------- */

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
ssize_t eibdrv_std_write(struct file *eib_filp,const char *buf,size_t len,loff_t *pos)
#else
read_write_t eibdrv_std_write(struct inode *eib_inodp,struct file *eib_filp,const char *buf,int len)
#endif
{  int checksum,i,result;
   int udata_len,msg_len;
   unsigned char msg[EIB_BUF_ELEMENT_SIZE+7];

   if (len>23) udata_len=23; else udata_len=len;
   
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
   if (copy_from_user(&msg[5],buf,udata_len)!=0)
      return -EFAULT;
#else
   memcpy_fromfs(&msg[5],buf,udata_len);
#endif   

   if (((eib_filp->f_flags&O_NONBLOCK)==O_NONBLOCK) &&
        (check_lock(&sv_ready)==0))
      return -EAGAIN;
      
   lock(&sv_ready);

   msg_len=udata_len+7;
   msg[0]=msg[3]=0x68;
   msg[1]=msg[2]=(unsigned char)(udata_len+1);
   msg[4]=(FT_DIR<<7) | 0x40 | (ftstation.fcb<<5) | (FT_FCB_VALID<<4) | 0x03;
   checksum=msg[4];
   for (i=0;i<udata_len;i++)
      checksum+=msg[i+5];
   msg[5+udata_len]=(unsigned char) checksum % 256;
   msg[6+udata_len]=0x16;
   ftstation.fcb=1-ftstation.fcb;

   if ((result=repeat_write(&msg[0],msg_len,FT_STATE_WAIT_FOR_UDATA_ACK))==msg_len)
      result=udata_len;
   
   unlock(&sv_ready);
   
   return result;
}

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
ssize_t eibdrv_std_read(struct file *eib_filp,char *buf,size_t len,loff_t *pos)
#else
read_write_t eibdrv_std_read(struct inode *eib_inodp,struct file *eib_filp,char *buf,int len)
#endif
{  int result,i;
   int blockmode;
   int copylen;
   int udata_len;
   int checksum;
   int msg_fcb;
   unsigned char msg[EIB_BUF_ELEMENT_SIZE+7];
   

   if (!(eib_filp->f_flags & O_NONBLOCK)) blockmode=FT_PROC_BLOCKED;
   else                                   blockmode=FT_PROC_NONBLOCKED;
   
   if ((blockmode==FT_PROC_NONBLOCKED)&&(check_lock(&sv_ready)==0))
      return -EAGAIN;
   lock(&sv_ready);
   
   fs=get_fs();
   set_fs(KERNEL_DS);
   
   while (((result=immediate_read(&msg[0]))!=1) && (blockmode==FT_PROC_BLOCKED))
      interruptible_sleep_on(&ftstation.port.tty->read_wait);

   if (result==0)
   {  unlock(&sv_ready);
      return 0;
   }
   
   copylen=0;   
   switch(msg[0])
   {  case FT_VAR_LENGTH_MSG: // SEND_UDATA received
                    if (timeout_read(&msg[0],4)!=4) break;
                    if ((msg[0]!=msg[1]) || (msg[0]>24) || (msg[2]!=0x68) ||
                       ((msg[3]&0xDF)!=(unsigned char)(((1-FT_DIR)<<7) | 0x40 | (FT_FCB_VALID<<4) | 0x03)))
                       break;

                    msg_fcb=(int)((msg[3] & 0x20)>>5);
                    checksum=msg[3];
                    udata_len=msg[0]-1;

                    if (timeout_read(&msg[0],udata_len+2)!=udata_len+2) break;
                    for (i=0;i<udata_len;i++)
                       checksum+=msg[i];
                    if ((msg[udata_len]!=(unsigned char)(checksum%256)) || (msg[udata_len+1]!=0x16))
                       break;
                    if (msg_fcb==(1-ftstation.rec_fcb))   // msg not already received
                    {  
#ifdef DEBUG                       
                       printk(KERN_WARNING "eibdrv (%i): SEND_UDATA  ==>  neue Nachricht\n",c++);
#endif                       
                       ftstation.rec_fcb=1-ftstation.rec_fcb;
                       ftstation.rrepeat=0;
                       copylen=udata_len;
                       if (copylen>len) copylen=len;
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)                       
                       if (copy_to_user(buf,&msg[0],copylen)!=0)
                          return -EFAULT;
#else
                       memcpy_tofs(buf,&msg[0],copylen);                    
#endif                                                 
                    }
                    else 
                    {  
#ifdef DEBUG                       
                       printk(KERN_WARNING "eibdrv (%i): SEND_UDATA  ==>  bereits erhalten\n",c++);
#endif                       
                       if (ftstation.rrepeat!=-1)
                          ftstation.rrepeat++;
                    }
                    if (ftstation.rrepeat<3)
                    {  if (ftstation.stat.read[ftstation.rrepeat]<ULONG_MAX)
                          ftstation.stat.read[ftstation.rrepeat]++;
                       else
                          ftstation.stat.overflow=1;
                       if (ftstation.rrepeat>0)
                          ftstation.stat.read[ftstation.rrepeat-1]--;  
                    }
                    line_idle_delay();
                    msg[0]=FT_ACK;
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)                    
                    ftstation.port.filp->f_op->write(ftstation.port.filp,&msg[0],1,&ftstation.port.filp->f_pos);
#else
                    ftstation.port.filp->f_op->write(ftstation.port.inodp,ftstation.port.filp,&msg[0],1);
#endif                                        
                    do_gettimeofday(&ftstation.last_seract);
                    break;
                    
            case FT_FIXED_LENGTH_MSG:
                    if (timeout_read(&msg[0],3)!=3) break;
                    if ((msg[0]!=msg[1]) || (msg[2]!=0x16)) break;
                    switch(msg[0] & 0xDF)
                    {  case (0x40 | ((1-FT_DIR)<<7)): // SEND_RESET
#ifdef DEBUG                       
                               printk(KERN_WARNING "eibdrv (%i): SEND_RESET  ==>  erhalten!\n",c++);
#endif DEBUG                               
                               ftstation.fcb=1;
                               ftstation.rec_fcb=0;
                               ftstation.wrepeat=FT_REPEAT_LIMIT;
                               line_idle_delay();
                               msg[0]=FT_ACK;
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
                               ftstation.port.filp->f_op->write(ftstation.port.filp,&msg[0],1,
                                  &ftstation.port.filp->f_pos);
#else                               
                               ftstation.port.filp->f_op->write(ftstation.port.inodp,ftstation.port.filp,&msg[0],1);
#endif
                               do_gettimeofday(&ftstation.last_seract);
                               break;
                       case (0x49 | ((1-FT_DIR)<<7)): // REQ_STATUS
#ifdef DEBUG                       
                               printk(KERN_WARNING "eibdrv (%i): REQ_STATUS  ==>  erhalten!\n",c++);
#endif DEBUG                               
                               msg[0]=0x10;
                               msg[1]=msg[2]=(0x0B | (FT_DIR<<7));
                               msg[3]=0x16;
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)                               
                               ftstation.port.filp->f_op->write(ftstation.port.filp,&msg[0],4,
                                  &ftstation.port.filp->f_pos);
#else
                               ftstation.port.filp->f_op->write(ftstation.port.inodp,ftstation.port.filp,&msg[0],4);
#endif                                                                 
                               do_gettimeofday(&ftstation.last_seract);
                               break;
                    }
                    break;
   }
   unlock(&sv_ready);
   return copylen;
}

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
int eibdrv_std_fasync(int eib_fd,struct file *eib_filp,int flag)
{  return fasync_helper(eib_fd,eib_filp,flag,&(ftstation.port.tty->fasync));
}
#else
int eibdrv_std_fasync(struct inode *eib_inodp,struct file *eib_filp,int flag)
{  return fasync_helper(eib_inodp,eib_filp,flag,&(ftstation.port.tty->fasync));
}
#endif

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,23)
unsigned int eibdrv_std_poll(struct file *eib_filp,struct poll_table_struct *eib_polltable)
{  unsigned int mask;
   
   poll_wait(eib_filp,&sv_ready.wq,eib_polltable);
   poll_wait(eib_filp,&ftstation.port.tty->read_wait,eib_polltable);
   
   mask = ftstation.port.filp->f_op->poll(ftstation.port.filp,eib_polltable);

   if (!check_lock(&sv_ready))
      mask &= ~(unsigned int)(POLLOUT | POLLWRNORM | POLLIN | POLLRDNORM);
   
   return mask;
}
#else
int eibdrv_std_select(struct inode *eib_inodp,struct file *eib_filp,int mode,select_table *eib_selecttable)
{  switch(mode)
   {  case SEL_IN:
      case SEL_OUT:
              if ((ftstation.port.filp->f_op->select(ftstation.port.inodp,
                        ftstation.port.filp,mode,eib_selecttable)==1) &&
                  (check_lock(&sv_ser_read)>0)) return 1;
              select_wait(&sv_ready.wq,eib_selecttable);
              select_wait(&ftstation.port.tty->read_wait,eib_selecttable);
              break;
      case SEL_EX:
              break;
   }
   return 0;
}
#endif

struct file_operations eibdrv_std_fops =
   { 
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)
     eibdrv_llseek,        // llseek
#else
     eibdrv_lseek,         // lseek
#endif
             
     eibdrv_std_read,      // read
     eibdrv_std_write,     // write
     NULL,                 // readdir
     
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,23)     
     eibdrv_std_poll,      // poll
#else
     eibdrv_std_select,    // select
#endif
               
     eibdrv_ioctl,         // ioctl
     NULL,                 // mmap
     eibdrv_open,          // open
     
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)     
     NULL,                 // flush
#endif
          
     eibdrv_release,       // release
     NULL,                 // fsync
     eibdrv_std_fasync,    // fasync
     NULL,                 // check_media_change
     NULL,                 // revalidate
     
#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,0)     
     NULL,                 // lock
#endif     
   };



int init_module(void)
{  int result;
   struct file_operations *fops=NULL;

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,18)
   EXPORT_NO_SYMBOLS;
#else
   register_symtab(NULL);
#endif
      
   switch(mode)
   {  case FT_SERVERMODE: 
              fops=&eibdrv_fops;
              ftstation.state=FT_STATE_INIT;
              ftstation.servproc=0;
              ftstation.servfunc=0;
              break;
        
      case FT_STANDARDMODE:
              fops = &eibdrv_std_fops;
              ftstation.state=FT_STATE_READY;
              break;
   }

   result = register_chrdev(major,"eibdrv",fops);
   if (result < 0)
   {  if (!major) printk(KERN_ERR "eibdrv (%i): cannot get free major number\n",c++);
      else        printk(KERN_ERR "eibdrv (%i): cannot get major number %i\n",c++,major);
      
      return result;
   }
   if (major == 0)          
      major = result;

   ftstation.f_exclusive=FT_SHARED_RW;
   ftstation.pasync=NULL;

   ftstation.fcb=1;
   ftstation.rec_fcb=0;
   ftstation.exchange_timeout=((int)((FT_EXCHANGE_TIMEOUT*HZ)/baudrate))+1;
   ftstation.exchange_time=jiffies_to_timeval(ftstation.exchange_timeout);
   ftstation.line_idle_timeout_usec=((suseconds_t)((FT_LINE_IDLE_TIMEOUT*1000000)/baudrate))+1;
   ftstation.line_idle_timeout_sec=((size_t)(ftstation.line_idle_timeout_usec/1000))+1;
   
   reset_statistics(&ftstation.stat);

   if ((result=open_ftstation())!=0) goto fail_open_ftstation;

   printk(KERN_WARNING "eibdrv " EIBDRV_VERSION " (GPL) by Wolfgang Tumfart (tumfart@auto.tuwien.ac.at)\n");
   return 0;
   
   fail_open_ftstation: unregister_chrdev(major,"eibdrv");
                        return result;
                        
      
}

void cleanup_module(void)
{  int result;
     
   switch(mode)
   {  case FT_SERVERMODE:
              ftstation.state=FT_STATE_SHUTDOWN;
      
              wake_up_interruptible(&wq_ser_write);
              wake_up_interruptible(&ftstation.port.tty->read_wait);
 
              if (ftstation.servfunc>0)
                 while (ftstation.servproc>0)
                    interruptible_sleep_on(&wq_shutdown); 
              break;
      case FT_STANDARDMODE:  
              ftstation.state=FT_STATE_SHUTDOWN;
              break;
   }
 
   close_ftstation();     
   result=unregister_chrdev(major,"eibdrv");
   if (result == -EINVAL)
   {  printk(KERN_ERR "eibdrv (%i): cannot deregister eibdrv with major %i\n",c++,major);
      return;
   }
   
   printk(KERN_WARNING "eibdrv " EIBDRV_VERSION " unloaded\n");
}


